@@ -30,13 +30,11 @@
 #include <stdint.h>
 #include <stdio.h>
 #include <string.h>
-#include <assert.h>
 #include <fcntl.h>
 #include <inttypes.h>
 #include <errno.h>
 #include <sys/stat.h>
 #include <sys/ioctl.h>
-#include <sys/mman.h>
 #include <sys/time.h>
 #include "drm.h"
 #include "i915_drm.h"
@@ -66,14 +64,15 @@ static int gem_linear_blt(uint32_t *batch,
 	uint32_t *b = batch;
 	int height = length / (16 * 1024);
 
-	assert(height <= 1<<16);
+	igt_assert(height <= 1<<16);
 
 	if (height) {
-		b[0] = COPY_BLT_CMD | BLT_WRITE_ALPHA | BLT_WRITE_RGB;
-		b[1] = 0xcc << 16 | 1 << 25 | 1 << 24 | (16*1024);
-		b[2] = 0;
-		b[3] = height << 16 | (4*1024);
-		b[4] = 0;
+		int i = 0;
+		b[i++] = COPY_BLT_CMD | BLT_WRITE_ALPHA | BLT_WRITE_RGB;
+		b[i++] = 0xcc << 16 | 1 << 25 | 1 << 24 | (16*1024);
+		b[i++] = 0;
+		b[i++] = height << 16 | (4*1024);
+		b[i++] = 0;
 		reloc->offset = (b-batch+4) * sizeof(uint32_t);
 		reloc->delta = 0;
 		reloc->target_handle = dst;
@@ -82,9 +81,9 @@ static int gem_linear_blt(uint32_t *batch,
 		reloc->presumed_offset = 0;
 		reloc++;
 
-		b[5] = 0;
-		b[6] = 16*1024;
-		b[7] = 0;
+		b[i++] = 0;
+		b[i++] = 16*1024;
+		b[i++] = 0;
 		reloc->offset = (b-batch+7) * sizeof(uint32_t);
 		reloc->delta = 0;
 		reloc->target_handle = src;
@@ -98,11 +97,12 @@ static int gem_linear_blt(uint32_t *batch,
 	}
 	
 	if (length) {
-		b[0] = COPY_BLT_CMD | BLT_WRITE_ALPHA | BLT_WRITE_RGB;
-		b[1] = 0xcc << 16 | 1 << 25 | 1 << 24 | (16*1024);
-		b[2] = height << 16;
-		b[3] = (1+height) << 16 | (length / 4);
-		b[4] = 0;
+		int i = 0;
+		b[i++] = COPY_BLT_CMD | BLT_WRITE_ALPHA | BLT_WRITE_RGB;
+		b[i++] = 0xcc << 16 | 1 << 25 | 1 << 24 | (16*1024);
+		b[i++] = height << 16;
+		b[i++] = (1+height) << 16 | (length / 4);
+		b[i++] = 0;
 		reloc->offset = (b-batch+4) * sizeof(uint32_t);
 		reloc->delta = 0;
 		reloc->target_handle = dst;
@@ -111,9 +111,9 @@ static int gem_linear_blt(uint32_t *batch,
 		reloc->presumed_offset = 0;
 		reloc++;
 
-		b[5] = height << 16;
-		b[6] = 16*1024;
-		b[7] = 0;
+		b[i++] = height << 16;
+		b[i++] = 16*1024;
+		b[i++] = 0;
 		reloc->offset = (b-batch+7) * sizeof(uint32_t);
 		reloc->delta = 0;
 		reloc->target_handle = src;
@@ -131,16 +131,6 @@ static int gem_linear_blt(uint32_t *batch,
 	return (b+2 - batch) * sizeof(uint32_t);
 }
 
-static void gem_exec(int fd, struct drm_i915_gem_execbuffer2 *execbuf)
-{
-	int ret;
-
-	ret = drmIoctl(fd,
-		       DRM_IOCTL_I915_GEM_EXECBUFFER2,
-		       execbuf);
-	assert(ret == 0);
-}
-
 static void run(int object_size)
 {
 	struct drm_i915_gem_execbuffer2 execbuf;
@@ -182,7 +172,7 @@ static void run(int object_size)
 	gem_write(fd, handle_relocs, 0, reloc, sizeof(reloc));
 	gtt_relocs = gem_mmap(fd, handle_relocs, 4096,
 			      PROT_READ | PROT_WRITE);
-	assert(gtt_relocs);
+	igt_assert(gtt_relocs);
 
 	exec[2].handle = handle;
 	exec[2].relocation_count = len > 40 ? 4 : 2;
@@ -210,7 +200,7 @@ static void run(int object_size)
 	i915_execbuffer2_set_context_id(execbuf, 0);
 	execbuf.rsvd2 = 0;
 
-	gem_exec(fd, &execbuf);
+	gem_execbuf(fd, &execbuf);
 	gem_sync(fd, handle);
 
 	gem_close(fd, handle);
@@ -218,9 +208,13 @@ static void run(int object_size)
 	close(fd);
 }
 
-int main(int argc, char **argv)
+igt_main
 {
-	run(OBJECT_SIZE);
-
-	return 0;
+	igt_subtest("normal")
+		run(OBJECT_SIZE);
+	igt_subtest("no-prefault") {
+		igt_disable_prefault();
+		run(OBJECT_SIZE);
+		igt_enable_prefault();
+	}
 }
