@@ -45,18 +45,22 @@
  *  - DP commands (e.g. poweroff)
  * - verify outputs against VBT/physical connectors
  */
+#ifdef HAVE_CONFIG_H
 #include "config.h"
+#endif
 
-#include <assert.h>
 #include <cairo.h>
 #include <errno.h>
 #include <math.h>
 #include <stdint.h>
+#include <stdbool.h>
+#include <strings.h>
 #include <unistd.h>
 #include <sys/poll.h>
 #include <sys/time.h>
-#include <sys/mman.h>
 #include <sys/ioctl.h>
+#include <sys/types.h>
+#include <sys/stat.h>
 
 #include "i915_drm.h"
 #include "drmtest.h"
@@ -67,11 +71,12 @@
 
 drmModeRes *resources;
 int drm_fd, modes;
-int dump_info = 0, test_all_modes =0, test_preferred_mode = 0, force_mode = 0,
-	test_plane, enable_tiling;
+int test_all_modes = 0, test_preferred_mode = 0, force_mode = 0, test_plane,
+    test_stereo_modes, enable_tiling;
 int sleep_between_modes = 5;
 uint32_t depth = 24, stride, bpp;
 int qr_code = 0;
+int specified_mode_num = -1, specified_disp_id = -1;
 
 drmModeModeInfo force_timing;
 
@@ -81,62 +86,6 @@ unsigned int plane_crtc_id;
 unsigned int plane_id;
 int plane_width, plane_height;
 static const uint32_t SPRITE_COLOR_KEY = 0x00aaaaaa;
-uint32_t *fb_ptr;
-
-#define ARRAY_SIZE(arr) (sizeof(arr) / sizeof((arr)[0]))
-
-struct type_name {
-	int type;
-	const char *name;
-};
-
-#define type_name_fn(res) \
-static const char * res##_str(int type) {			\
-	unsigned int i;					\
-	for (i = 0; i < ARRAY_SIZE(res##_names); i++) { \
-		if (res##_names[i].type == type)	\
-			return res##_names[i].name;	\
-	}						\
-	return "(invalid)";				\
-}
-
-struct type_name encoder_type_names[] = {
-	{ DRM_MODE_ENCODER_NONE, "none" },
-	{ DRM_MODE_ENCODER_DAC, "DAC" },
-	{ DRM_MODE_ENCODER_TMDS, "TMDS" },
-	{ DRM_MODE_ENCODER_LVDS, "LVDS" },
-	{ DRM_MODE_ENCODER_TVDAC, "TVDAC" },
-};
-
-type_name_fn(encoder_type)
-
-struct type_name connector_status_names[] = {
-	{ DRM_MODE_CONNECTED, "connected" },
-	{ DRM_MODE_DISCONNECTED, "disconnected" },
-	{ DRM_MODE_UNKNOWNCONNECTION, "unknown" },
-};
-
-type_name_fn(connector_status)
-
-struct type_name connector_type_names[] = {
-	{ DRM_MODE_CONNECTOR_Unknown, "unknown" },
-	{ DRM_MODE_CONNECTOR_VGA, "VGA" },
-	{ DRM_MODE_CONNECTOR_DVII, "DVI-I" },
-	{ DRM_MODE_CONNECTOR_DVID, "DVI-D" },
-	{ DRM_MODE_CONNECTOR_DVIA, "DVI-A" },
-	{ DRM_MODE_CONNECTOR_Composite, "composite" },
-	{ DRM_MODE_CONNECTOR_SVIDEO, "s-video" },
-	{ DRM_MODE_CONNECTOR_LVDS, "LVDS" },
-	{ DRM_MODE_CONNECTOR_Component, "component" },
-	{ DRM_MODE_CONNECTOR_9PinDIN, "9-pin DIN" },
-	{ DRM_MODE_CONNECTOR_DisplayPort, "DisplayPort" },
-	{ DRM_MODE_CONNECTOR_HDMIA, "HDMI-A" },
-	{ DRM_MODE_CONNECTOR_HDMIB, "HDMI-B" },
-	{ DRM_MODE_CONNECTOR_TV, "TV" },
-	{ DRM_MODE_CONNECTOR_eDP, "Embedded DisplayPort" },
-};
-
-type_name_fn(connector_type)
 
 /*
  * Mode setting with the kernel interfaces is a bit of a chore.
@@ -152,6 +101,7 @@ struct connector {
 	drmModeEncoder *encoder;
 	drmModeConnector *connector;
 	int crtc;
+	int crtc_idx;
 	int pipe;
 };
 
@@ -182,8 +132,8 @@ static void dump_connectors_fd(int drmfd)
 		printf("%d\t%d\t%s\t%s\t%dx%d\t\t%d\n",
 		       connector->connector_id,
 		       connector->encoder_id,
-		       connector_status_str(connector->connection),
-		       connector_type_str(connector->connector_type),
+		       kmstest_connector_status_str(connector->connection),
+		       kmstest_connector_type_str(connector->connector_type),
 		       connector->mmWidth, connector->mmHeight,
 		       connector->count_modes);
 
@@ -193,8 +143,10 @@ static void dump_connectors_fd(int drmfd)
 		printf("  modes:\n");
 		printf("  name refresh (Hz) hdisp hss hse htot vdisp "
 		       "vss vse vtot flags type clock\n");
-		for (j = 0; j < connector->count_modes; j++)
+		for (j = 0; j < connector->count_modes; j++){
+			fprintf(stdout, "[%d]", j );
 			kmstest_dump_mode(&connector->modes[j]);
+		}
 
 		drmModeFreeConnector(connector);
 	}
@@ -233,109 +185,58 @@ static void dump_crtcs_fd(int drmfd)
 	drmModeFreeResources(mode_resources);
 }
 
-static void connector_find_preferred_mode(struct connector *c)
+static void dump_info(void)
 {
-	drmModeConnector *connector;
-	drmModeEncoder *encoder = NULL;
-	int i, j;
+	dump_connectors_fd(drm_fd);
+	dump_crtcs_fd(drm_fd);
+}
 
-	/* First, find the connector & mode */
-	c->mode_valid = 0;
-	connector = drmModeGetConnector(drm_fd, c->id);
-	if (!connector) {
-		fprintf(stderr, "could not get connector %d: %s\n",
-			c->id, strerror(errno));
-		drmModeFreeConnector(connector);
-		return;
-	}
+static void connector_find_preferred_mode(uint32_t connector_id,
+					  unsigned long crtc_idx_mask,
+					  int mode_num, struct connector *c)
+{
+	struct kmstest_connector_config config;
 
-	if (connector->connection != DRM_MODE_CONNECTED) {
-		drmModeFreeConnector(connector);
-		return;
-	}
-
-	if (!connector->count_modes) {
-		fprintf(stderr, "connector %d has no modes\n", c->id);
-		drmModeFreeConnector(connector);
-		return;
-	}
-
-	if (connector->connector_id != c->id) {
-		fprintf(stderr, "connector id doesn't match (%d != %d)\n",
-			connector->connector_id, c->id);
-		drmModeFreeConnector(connector);
-		return;
-	}
-
-	for (j = 0; j < connector->count_modes; j++) {
-		c->mode = connector->modes[j];
-		if (c->mode.type & DRM_MODE_TYPE_PREFERRED) {
-			c->mode_valid = 1;
-			break;
-		}
-	}
-
-	if (!c->mode_valid) {
-		if (connector->count_modes > 0) {
-			/* use the first mode as test mode */
-			c->mode = connector->modes[0];
-			c->mode_valid = 1;
-		}
-		else {
-			fprintf(stderr, "failed to find any modes on connector %d\n",
-				c->id);
-			return;
-		}
-	}
-
-	/* Now get the encoder */
-	for (i = 0; i < connector->count_encoders; i++) {
-		encoder = drmModeGetEncoder(drm_fd, connector->encoders[i]);
-
-		if (!encoder) {
-			fprintf(stderr, "could not get encoder %i: %s\n",
-				resources->encoders[i], strerror(errno));
-			drmModeFreeEncoder(encoder);
-			continue;
-		}
-
-		break;
-	}
-
-	c->encoder = encoder;
-
-	if (i == resources->count_encoders) {
-		fprintf(stderr, "failed to find encoder\n");
+	if (kmstest_get_connector_config(drm_fd, connector_id, crtc_idx_mask,
+					 &config) < 0) {
 		c->mode_valid = 0;
 		return;
 	}
 
-	/* Find first CRTC not in use */
-	for (i = 0; i < resources->count_crtcs; i++) {
-		if (resources->crtcs[i] && (c->encoder->possible_crtcs & (1<<i)))
-			break;
+	c->connector = config.connector;
+	c->encoder = config.encoder;
+	c->crtc = config.crtc->crtc_id;
+	c->crtc_idx = config.crtc_idx;
+	c->pipe = config.pipe;
+
+	if (mode_num != -1) {
+		igt_assert(mode_num < config.connector->count_modes);
+		c->mode = config.connector->modes[mode_num];
+	} else {
+		c->mode = config.default_mode;
 	}
-	c->crtc = resources->crtcs[i];
-	c->pipe = i;
-
-	if(test_preferred_mode || force_mode)
-		resources->crtcs[i] = 0;
-
-	c->connector = connector;
+	c->mode_valid = 1;
 }
 
 static void
-paint_color_key(void)
+paint_color_key(struct kmstest_fb *fb_info)
 {
 	int i, j;
+	uint32_t *fb_ptr;
+
+	fb_ptr = gem_mmap(drm_fd, fb_info->gem_handle,
+			  fb_info->size, PROT_READ | PROT_WRITE);
+	igt_assert(fb_ptr);
 
 	for (i = crtc_y; i < crtc_y + crtc_h; i++)
 		for (j = crtc_x; j < crtc_x + crtc_w; j++) {
 			uint32_t offset;
 
-			offset = (i * width) + j;
+			offset = (i * fb_info->stride / 4) + j;
 			fb_ptr[offset] = SPRITE_COLOR_KEY;
 		}
+
+	munmap(fb_ptr, fb_info->size);
 }
 
 static void paint_image(cairo_t *cr, const char *file)
@@ -369,94 +270,61 @@ static void paint_image(cairo_t *cr, const char *file)
 	cairo_surface_destroy(image);
 }
 
-static void
-paint_output_info(cairo_t *cr, int l_width, int l_height, void *priv)
+static void paint_output_info(struct connector *c, struct kmstest_fb *fb)
 {
-	struct connector *c = priv;
-	cairo_text_extents_t name_extents, mode_extents;
-	char name_buf[128], mode_buf[128];
-	int i, x, y, modes_x, modes_y;
+	cairo_t *cr = kmstest_get_cairo_ctx(drm_fd, fb);
+	int l_width = fb->width;
+	int l_height = fb->height;
+	double str_width;
+	double x, y, top_y;
+	double max_width;
+	int i;
+
+	kmstest_paint_test_pattern(cr, l_width, l_height);
 
-	/* Get text extents for each string */
-	snprintf(name_buf, sizeof name_buf, "%s",
-		 connector_type_str(c->connector->connector_type));
-	cairo_set_font_size(cr, 48);
 	cairo_select_font_face(cr, "Helvetica",
 			       CAIRO_FONT_SLANT_NORMAL,
 			       CAIRO_FONT_WEIGHT_NORMAL);
-	cairo_text_extents(cr, name_buf, &name_extents);
+	cairo_move_to(cr, l_width / 2, l_height / 2);
 
-	snprintf(mode_buf, sizeof mode_buf, "%s @ %dHz on %s encoder",
-		 c->mode.name, c->mode.vrefresh,
-		 encoder_type_str(c->encoder->encoder_type));
-	cairo_set_font_size(cr, 36);
-	cairo_text_extents(cr, mode_buf, &mode_extents);
-
-	/* Paint output name */
-	x = l_width / 2;
-	x -= name_extents.width / 2;
-	y = l_height / 2;
-	y -= (name_extents.height / 2) - (mode_extents.height / 2) - 10;
+	/* Print connector and mode name */
 	cairo_set_font_size(cr, 48);
-	cairo_move_to(cr, x, y);
-	cairo_text_path(cr, name_buf);
-	cairo_set_source_rgb(cr, 0, 0, 0);
-	cairo_stroke_preserve(cr);
-	cairo_set_source_rgb(cr, 1, 1, 1);
-	cairo_fill(cr);
+	kmstest_cairo_printf_line(cr, align_hcenter, 10, "%s",
+		 kmstest_connector_type_str(c->connector->connector_type));
 
-	/* Paint mode name */
-	x = l_width / 2;
-	x -= mode_extents.width / 2;
-	modes_x = x;
-	y = l_height / 2;
-	y += (mode_extents.height / 2) + (name_extents.height / 2) + 10;
 	cairo_set_font_size(cr, 36);
-	cairo_move_to(cr, x, y);
-	cairo_text_path(cr, mode_buf);
-	cairo_set_source_rgb(cr, 0, 0, 0);
-	cairo_stroke_preserve(cr);
-	cairo_set_source_rgb(cr, 1, 1, 1);
-	cairo_fill(cr);
+	str_width = kmstest_cairo_printf_line(cr, align_hcenter, 10,
+		"%s @ %dHz on %s encoder", c->mode.name, c->mode.vrefresh,
+		kmstest_encoder_type_str(c->encoder->encoder_type));
+
+	cairo_rel_move_to(cr, -str_width / 2, 0);
 
 	/* List available modes */
-	snprintf(mode_buf, sizeof mode_buf, "Available modes:");
 	cairo_set_font_size(cr, 18);
-	cairo_text_extents(cr, mode_buf, &mode_extents);
-	x = modes_x;
-	modes_x = x + mode_extents.width;
-	y += mode_extents.height + 10;
-	modes_y = y;
-	cairo_move_to(cr, x, y);
-	cairo_text_path(cr, mode_buf);
-	cairo_set_source_rgb(cr, 0, 0, 0);
-	cairo_stroke_preserve(cr);
-	cairo_set_source_rgb(cr, 1, 1, 1);
-	cairo_fill(cr);
+	str_width = kmstest_cairo_printf_line(cr, align_left, 10,
+					      "Available modes:");
+	cairo_rel_move_to(cr, str_width, 0);
+	cairo_get_current_point(cr, &x, &top_y);
 
+	max_width = 0;
 	for (i = 0; i < c->connector->count_modes; i++) {
-		snprintf(mode_buf, sizeof mode_buf, "%s @ %dHz",
-			 c->connector->modes[i].name,
-			 c->connector->modes[i].vrefresh);
-		cairo_set_font_size(cr, 18);
-		cairo_text_extents(cr, mode_buf, &mode_extents);
-		x = modes_x - mode_extents.width; /* right justify modes */
-		y += mode_extents.height + 10;
-		if (y + mode_extents.height >= height) {
-			y = modes_y + mode_extents.height + 10;
-			modes_x += mode_extents.width + 10;
-			x = modes_x - mode_extents.width;
+		cairo_get_current_point(cr, &x, &y);
+		if (y >= l_height) {
+			x += max_width + 10;
+			max_width = 0;
+			cairo_move_to(cr, x, top_y);
 		}
-		cairo_move_to(cr, x, y);
-		cairo_text_path(cr, mode_buf);
-		cairo_set_source_rgb(cr, 0, 0, 0);
-		cairo_stroke_preserve(cr);
-		cairo_set_source_rgb(cr, 1, 1, 1);
-		cairo_fill(cr);
+		str_width = kmstest_cairo_printf_line(cr, align_right, 10,
+			"%s @ %dHz", c->connector->modes[i].name,
+			 c->connector->modes[i].vrefresh);
+		if (str_width > max_width)
+			max_width = str_width;
 	}
 
 	if (qr_code)
-		paint_image(cr, "./pass.png");
+		paint_image(cr, IGT_DATADIR"/pass.png");
+
+	igt_assert(!cairo_status(cr));
 }
 
 static void sighandler(int signo)
@@ -480,18 +348,8 @@ static void
 set_mode(struct connector *c)
 {
 	unsigned int fb_id = 0;
-	int j, test_mode_num;
-
-	if (depth <= 8)
-		bpp = 8;
-	else if (depth > 8 && depth <= 16)
-		bpp = 16;
-	else if (depth > 16 && depth <= 32)
-		bpp = 32;
-
-	connector_find_preferred_mode(c);
-	if (!c->mode_valid)
-		return;
+	struct kmstest_fb fb_info[2] = { };
+	int j, test_mode_num, current_fb = 0, old_fb = -1;
 
 	test_mode_num = 1;
 	if (force_mode){
@@ -503,11 +361,14 @@ set_mode(struct connector *c)
 		test_mode_num = c->connector->count_modes;
 
 	for (j = 0; j < test_mode_num; j++) {
-		struct kmstest_fb fb_info;
 
 		if (test_all_modes)
 			c->mode = c->connector->modes[j];
 
+		/* set_mode() only tests 2D modes */
+		if (c->mode.flags & DRM_MODE_FLAG_3D_MASK)
+			continue;
+
 		if (!c->mode_valid)
 			continue;
 
@@ -515,17 +376,11 @@ set_mode(struct connector *c)
 		height = c->mode.vdisplay;
 
 		fb_id = kmstest_create_fb(drm_fd, width, height, bpp, depth,
-					  enable_tiling, &fb_info,
-					  paint_output_info, c);
+					  enable_tiling, &fb_info[current_fb]);
+		paint_output_info(c, &fb_info[current_fb]);
+		paint_color_key(&fb_info[current_fb]);
 
-		fb_ptr = gem_mmap(drm_fd, fb_info.gem_handle,
-				  fb_info.size, PROT_READ | PROT_WRITE);
-		assert(fb_ptr);
-		paint_color_key();
-
-		gem_close(drm_fd, fb_info.gem_handle);
-
-		fprintf(stdout, "CRTS(%u):",c->crtc);
+		fprintf(stdout, "CRTC(%u):[%d]",c->crtc, j);
 		kmstest_dump_mode(&c->mode);
 		if (drmModeSetCrtc(drm_fd, c->crtc, fb_id, 0, 0,
 				   &c->id, 1, &c->mode)) {
@@ -535,6 +390,11 @@ set_mode(struct connector *c)
 			continue;
 		}
 
+		if (old_fb != -1)
+			kmstest_remove_fb(drm_fd, &fb_info[old_fb]);
+		old_fb = current_fb;
+		current_fb = 1 - current_fb;
+
 		if (sleep_between_modes && test_all_modes && !qr_code)
 			sleep(sleep_between_modes);
 
@@ -542,12 +402,170 @@ set_mode(struct connector *c)
 			set_single();
 			pause();
 		}
-
 	}
 
-	if(test_all_modes){
-		drmModeRmFB(drm_fd,fb_id);
-		drmModeSetCrtc(drm_fd, c->crtc, fb_id, 0, 0,  &c->id, 1, 0);
+	if (test_all_modes)
+		kmstest_remove_fb(drm_fd, &fb_info[old_fb]);
+
+	drmModeFreeEncoder(c->encoder);
+	drmModeFreeConnector(c->connector);
+}
+
+struct box {
+	int x, y, width, height;
+};
+
+struct stereo_fb_layout {
+	int fb_width, fb_height;
+	struct box left, right;
+};
+
+static void box_init(struct box *box, int x, int y, int bwidth, int bheight)
+{
+	box->x = x;
+	box->y = y;
+	box->width = bwidth;
+	box->height = bheight;
+}
+
+static void stereo_fb_layout_from_mode(struct stereo_fb_layout *layout,
+				       drmModeModeInfo *mode)
+{
+	unsigned int format = mode->flags & DRM_MODE_FLAG_3D_MASK;
+	const int hdisplay = mode->hdisplay, vdisplay = mode->vdisplay;
+	int middle;
+
+	switch (format) {
+	case DRM_MODE_FLAG_3D_TOP_AND_BOTTOM:
+		layout->fb_width = hdisplay;
+		layout->fb_height = vdisplay;
+
+		middle = vdisplay / 2;
+		box_init(&layout->left, 0, 0, hdisplay, middle);
+		box_init(&layout->right,
+			 0, middle, hdisplay, vdisplay - middle);
+		break;
+	case DRM_MODE_FLAG_3D_SIDE_BY_SIDE_HALF:
+		layout->fb_width = hdisplay;
+		layout->fb_height = vdisplay;
+
+		middle = hdisplay / 2;
+		box_init(&layout->left, 0, 0, middle, vdisplay);
+		box_init(&layout->right,
+			 middle, 0, hdisplay - middle, vdisplay);
+		break;
+	case DRM_MODE_FLAG_3D_FRAME_PACKING:
+	{
+		int vactive_space = mode->vtotal - vdisplay;
+
+		layout->fb_width = hdisplay;
+		layout->fb_height = 2 * vdisplay + vactive_space;
+
+		box_init(&layout->left,
+			 0, 0, hdisplay, vdisplay);
+		box_init(&layout->right,
+			 0, vdisplay + vactive_space, hdisplay, vdisplay);
+		break;
+	}
+	default:
+		assert(0);
+	}
+}
+
+static const char *stereo_mode_str(drmModeModeInfo *mode)
+{
+	unsigned int layout = mode->flags & DRM_MODE_FLAG_3D_MASK;
+
+	switch (layout) {
+	case DRM_MODE_FLAG_3D_TOP_AND_BOTTOM:
+		return "TB";
+	case DRM_MODE_FLAG_3D_SIDE_BY_SIDE_HALF:
+		return "SbSH";
+	case DRM_MODE_FLAG_3D_FRAME_PACKING:
+		return "FP";
+	default:
+		assert(0);
+	}
+}
+
+static uint32_t create_stereo_fb(drmModeModeInfo *mode, struct kmstest_fb *fb)
+{
+	struct stereo_fb_layout layout;
+	cairo_t *cr;
+	uint32_t fb_id;
+
+	stereo_fb_layout_from_mode(&layout, mode);
+	fb_id = kmstest_create_fb(drm_fd, layout.fb_width, layout.fb_height,
+				  bpp, depth, enable_tiling, fb);
+	cr = kmstest_get_cairo_ctx(drm_fd, fb);
+
+	kmstest_paint_image(cr, IGT_DATADIR"/1080p-left.png",
+			    layout.left.x, layout.left.y,
+			    layout.left.width, layout.left.height);
+	kmstest_paint_image(cr, IGT_DATADIR"/1080p-right.png",
+			    layout.right.x, layout.right.y,
+			    layout.right.width, layout.right.height);
+
+	{
+		char buffer[64];
+
+		snprintf(buffer, sizeof(buffer), "%dx%d@%dHz-%s.png",
+			 mode->hdisplay,
+			 mode->vdisplay,
+			 mode->vrefresh,
+			 stereo_mode_str(mode));
+
+		kmstest_write_fb(drm_fd, fb, buffer);
+	}
+
+	return fb_id;
+}
+
+static void do_set_stereo_mode(struct connector *c)
+{
+	uint32_t fb_id;
+	struct kmstest_fb fb_info;
+
+	fb_id = create_stereo_fb(&c->mode, &fb_info);
+
+	if (drmModeSetCrtc(drm_fd, c->crtc, fb_id, 0, 0,
+			   &c->id, 1, &c->mode)) {
+		fprintf(stderr, "failed to set mode (%dx%d@%dHz): %s\n",
+			width, height, c->mode.vrefresh,
+			strerror(errno));
+	}
+}
+
+static void
+set_stereo_mode(struct connector *c)
+{
+	int i, n;
+
+
+	if (specified_mode_num != -1)
+		n = 1;
+	else
+		n = c->connector->count_modes;
+
+	for (i = 0; i < n; i++) {
+		if (specified_mode_num == -1)
+			c->mode = c->connector->modes[i];
+
+		if (!c->mode_valid)
+			continue;
+
+		if (!(c->mode.flags & DRM_MODE_FLAG_3D_MASK))
+			continue;
+
+		fprintf(stdout, "CRTC(%u): [%d]", c->crtc, i);
+		kmstest_dump_mode(&c->mode);
+		do_set_stereo_mode(c);
+
+		if (qr_code) {
+			set_single();
+			pause();
+		} else if (sleep_between_modes)
+			sleep(sleep_between_modes);
 	}
 
 	drmModeFreeEncoder(c->encoder);
@@ -580,25 +598,63 @@ int update_display(void)
 	if (!connectors)
 		return 0;
 
-	if (dump_info) {
-		dump_connectors_fd(drm_fd);
-		dump_crtcs_fd(drm_fd);
-	}
+	if (test_preferred_mode || test_all_modes ||
+	    force_mode || specified_disp_id != -1) {
+		unsigned long crtc_idx_mask = -1UL;
 
-	if (test_preferred_mode || test_all_modes || force_mode) {
 		/* Find any connected displays */
 		for (c = 0; c < resources->count_connectors; c++) {
-			connectors[c].id = resources->connectors[c];
-			set_mode(&connectors[c]);
+			struct connector *connector = &connectors[c];
+
+			connector->id = resources->connectors[c];
+			if (specified_disp_id != -1 &&
+			    connector->id != specified_disp_id)
+				continue;
+
+			connector_find_preferred_mode(connector->id,
+						      crtc_idx_mask,
+						      specified_mode_num,
+						      connector);
+			if (!connector->mode_valid)
+				continue;
+
+			set_mode(connector);
+
+			if (test_preferred_mode || force_mode ||
+			    specified_mode_num != -1)
+				crtc_idx_mask &= ~(1 << connector->crtc_idx);
+
 		}
 	}
+
+	if (test_stereo_modes) {
+		for (c = 0; c < resources->count_connectors; c++) {
+			struct connector *connector = &connectors[c];
+
+			connector->id = resources->connectors[c];
+			if (specified_disp_id != -1 &&
+			    connector->id != specified_disp_id)
+				continue;
+
+			connector_find_preferred_mode(connector->id,
+						      -1UL,
+						      specified_mode_num,
+						      connector);
+			if (!connector->mode_valid)
+				continue;
+
+			set_stereo_mode(connector);
+		}
+	}
+
+	free(connectors);
 	drmModeFreeResources(resources);
 	return 1;
 }
 
-static char optstr[] = "hiaf:s:d:p:mrt";
+static char optstr[] = "3hiaf:s:d:p:mrto:";
 
-static void usage(char *name)
+static void __attribute__((noreturn)) usage(char *name)
 {
 	fprintf(stderr, "usage: %s [-hiasdpmtf]\n", name);
 	fprintf(stderr, "\t-i\tdump info\n");
@@ -607,8 +663,10 @@ static void usage(char *name)
 	fprintf(stderr, "\t-d\t<depth>\tbit depth of scanout buffer\n");
 	fprintf(stderr, "\t-p\t<planew,h>,<crtcx,y>,<crtcw,h> test overlay plane\n");
 	fprintf(stderr, "\t-m\ttest the preferred mode\n");
+	fprintf(stderr, "\t-3\ttest all 3D modes\n");
 	fprintf(stderr, "\t-t\tuse a tiled framebuffer\n");
 	fprintf(stderr, "\t-r\tprint a QR code on the screen whose content is \"pass\" for the automatic test\n");
+	fprintf(stderr, "\t-o\t<id of the display>,<number of the mode>\tonly test specified mode on the specified display\n");
 	fprintf(stderr, "\t-f\t<clock MHz>,<hdisp>,<hsync-start>,<hsync-end>,<htotal>,\n");
 	fprintf(stderr, "\t\t<vdisp>,<vsync-start>,<vsync-end>,<vtotal>\n");
 	fprintf(stderr, "\t\ttest force mode\n");
@@ -637,6 +695,7 @@ static void enter_exec_path( char **argv )
 	char *exec_path = NULL;
 	char *pos = NULL;
 	short len_path = 0;
+	int ret;
 
 	len_path = strlen( argv[0] );
 	exec_path = (char*) malloc(len_path);
@@ -646,7 +705,8 @@ static void enter_exec_path( char **argv )
 	if (pos != NULL)
 		*(pos+1) = '\0';
 
-	chdir(exec_path);
+	ret = chdir(exec_path);
+	igt_assert(ret == 0);
 	free(exec_path);
 }
 
@@ -657,14 +717,19 @@ int main(int argc, char **argv)
 	GIOChannel *stdinchannel;
 	GMainLoop *mainloop;
 	float force_clock;
+	bool opt_dump_info = false;
+
+	igt_skip_on_simulation();
 
 	enter_exec_path( argv );
 
-	opterr = 0;
 	while ((c = getopt(argc, argv, optstr)) != -1) {
 		switch (c) {
+		case '3':
+			test_stereo_modes = 1;
+			break;
 		case 'i':
-			dump_info = 1;
+			opt_dump_info = true;
 			break;
 		case 'a':
 			test_all_modes = 1;
@@ -701,20 +766,43 @@ int main(int argc, char **argv)
 		case 'r':
 			qr_code = 1;
 			break;
+		case 'o':
+			sscanf(optarg, "%d,%d", &specified_disp_id, &specified_mode_num);
+			break;
 		default:
-			fprintf(stderr, "unknown option %c\n", c);
 			/* fall through */
 		case 'h':
 			usage(argv[0]);
 			break;
 		}
 	}
-	if (!test_all_modes && !force_mode && !dump_info &&
-	    !test_preferred_mode)
+
+	if (depth <= 8)
+		bpp = 8;
+	else if (depth <= 16)
+		bpp = 16;
+	else if (depth <= 32)
+		bpp = 32;
+
+	if (!test_all_modes && !force_mode && !test_preferred_mode &&
+	    specified_mode_num == -1 && !test_stereo_modes)
 		test_all_modes = 1;
 
 	drm_fd = drm_open_any();
 
+	if (test_stereo_modes &&
+	    drmSetClientCap(drm_fd, DRM_CLIENT_CAP_STEREO_3D, 1) < 0) {
+		fprintf(stderr, "DRM_CLIENT_CAP_STEREO_3D failed\n");
+		goto out_close;
+	}
+
+	if (opt_dump_info) {
+		dump_info();
+		goto out_close;
+	}
+
+	igt_set_vt_graphics_mode();
+
 	mainloop = g_main_loop_new(NULL, FALSE);
 	if (!mainloop) {
 		fprintf(stderr, "failed to create glib mainloop\n");
@@ -747,7 +835,7 @@ int main(int argc, char **argv)
 		goto out_stdio;
 	}
 
-	if (dump_info || test_all_modes)
+	if (test_all_modes)
 		goto out_stdio;
 
 	g_main_loop_run(mainloop);
